/* os_win32.c - Windows specific implementation details of the os module

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include "aconfig.h"

#if defined(A_HAVE_OS_MODULE) && defined(A_HAVE_WINDOWS)

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <dirent.h>
#include <time.h>
#include <windows.h>
#include "alore.h"
#include "os_module.h"
#include "io_module.h"
#include "str.h"
#include "array.h"
#include "internal.h"
#include "runtime.h"


#define MAX_ENV_VALUE_LEN 4096


static AValue MakeDateTime(AThread *t, AValue *frame, time_t t1);


AValue ARemoveImpl(AThread *t, const AMixedCValue *frame)
{
    int status;

    AAllowBlocking();
    status = remove(frame[0].strValue);
    AEndBlocking();
    
    if (status < 0) {
        if (errno == EACCES || errno == ENOENT) {
            AAllowBlocking();
            status = RemoveDirectory(frame[0].strValue);
            AEndBlocking();
            
            if (!status) {
                errno = EACCES;
                return ARaiseErrnoIoError(t);
            } else
                return ANil;
        } else
            return ARaiseErrnoIoError(t);
    } else
        return ANil;
}


AValue ARenameImpl(AThread *t, const AMixedCValue *frame)
{
    int status;
    
    /* IDEA: Maybe return a special error if cannot rename because destination
       on a different file system. */

    /* IDEA: Probably create the directories beneath the destination directory
       as needed. */

    AAllowBlocking();
    status = rename(frame[0].strValue, frame[1].strValue);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue AMakeDirImpl(AThread *t, const AMixedCValue *frame)
{
    int status;
    
    /* IDEA: Create the directories beneath the destination directory as
       needed. */
    AAllowBlocking();
    status = mkdir(frame[0].strValue);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue AChangeDirImpl(AThread *t, const AMixedCValue *frame)
{
    int status;

    AAllowBlocking();
    status = chdir(frame[0].strValue);
    AEndBlocking();
    
    if (status < 0)
        return ARaiseErrnoIoError(t);
    else
        return ANil;
}


AValue ACurrentDirImpl(AThread *t, const AMixedCValue *frame)
{
    char buf[A_MAX_PATH_LEN];

    if (getcwd(buf, A_MAX_PATH_LEN) == NULL)
        return ARaiseErrnoIoError(t);
    else
        return AMakeStr(t, buf);
}


AValue ASleepImpl(AThread *t, const AMixedCValue *frame)
{
    int ms = (int)(frame[0].floatValue * 1000 + 0.5);
    if (ms == 0 && frame[0].floatValue > 0.0)
        ms = 1;
    AAllowBlocking();
    Sleep(ms);
    AEndBlocking();
    return ANil;
}


AValue ASystemImpl(AThread *t, const AMixedCValue *frame)
{
    /* IDEA: Can we analyze the return code and raise an exception or
       something -- but perhaps we'd better not. */
    int error;

    AAllowBlocking();
    error = system(frame[0].strValue);
    AEndBlocking();
    
    return AIntToValue(error);
}


AValue AExistsImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }
    
    return ATrue;
}


AValue AIsFileImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }
    
    return S_ISREG(buf.st_mode) ? ATrue : AFalse;
}


AValue AIsDirImpl(AThread *t, const AMixedCValue *frame)
{
    struct stat buf;
    int status;

    AAllowBlocking();
    status = stat(frame[0].strValue, &buf);
    AEndBlocking();

    if (status < 0) {
        if (errno == ENOENT || errno == ENOTDIR)
            return AFalse;
        return ARaiseErrnoIoError(t);
    }

    return S_ISDIR(buf.st_mode) ? ATrue : AFalse;
}


AValue AIsLinkImpl(AThread *t, const AMixedCValue *frame)
{
    /* IDEA: Detect shortcuts? */
    return AFalse;
}


enum {
    STAT_PATH,
    STAT_FILLER1, /* Posix-only */
    STAT_FILLER2, /* Posix-only */
    STAT_SIZE,
    STAT_IS_FILE,
    STAT_IS_DIR,
    STAT_IS_LINK,
    STAT_IS_SPECIAL
};


AValue AStatCreate(AThread *t, AValue *frame)
{
    char path[A_MAX_PATH_LEN];
    struct stat buf;
    ABool isLink = FALSE;
    int status;

    AGetStr(t, frame[1], path, A_MAX_PATH_LEN);

    AAllowBlocking();
    status = stat(path, &buf);
    AEndBlocking();

    if (status < 0)
        return ARaiseErrnoIoError(t);

    /* IDEA: Detect shortcuts? */
    isLink = FALSE;

    ASetMemberDirect(t, frame[0], STAT_PATH, frame[1]);
    /* FIX: Max size is 512MB!! */
    ASetMemberDirect(t, frame[0], STAT_SIZE, AIntToValue(buf.st_size));
    ASetMemberDirect(t, frame[0], STAT_IS_FILE, S_ISREG(buf.st_mode)
                    ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_DIR, S_ISDIR(buf.st_mode)
                    ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_LINK, isLink ? ATrue : AFalse);
    ASetMemberDirect(t, frame[0], STAT_IS_SPECIAL, S_ISCHR(buf.st_mode) ||
                    S_ISBLK(buf.st_mode) || S_ISFIFO(buf.st_mode)
                    ? ATrue : AFalse);

    *(time_t *)ADataPtr(frame[0], 0) = buf.st_mtime;
    *(time_t *)ADataPtr(frame[0], sizeof(time_t)) = buf.st_atime;
    
    return frame[0];
}


AValue AStatModificationTime(AThread *t, AValue *frame)
{
    time_t t1 = *(time_t *)ADataPtr(frame[0], 0);
    return MakeDateTime(t, frame, t1);
}


AValue AStatAccessTime(AThread *t, AValue *frame)
{
    time_t t1 = *(time_t *)ADataPtr(frame[0], sizeof(time_t));
    return MakeDateTime(t, frame, t1);
}


static AValue MakeDateTime(AThread *t, AValue *frame, time_t t1)
{
    struct tm t2;
    char buf[30];

    t2 = *localtime(&t1);
    
    sprintf(buf, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d",
            t2.tm_year + 1900,
            t2.tm_mon + 1,
            t2.tm_mday,
            t2.tm_hour,
            t2.tm_min,
            AMin(59, t2.tm_sec));

    return AMakeStr(t, buf);
}


AValue AStatIsReadable(AThread *t, AValue *frame)
{
    AValue pathVal = AMemberDirect(frame[0], 0);
    char path[A_MAX_PATH_LEN];
    int status;
    
    AGetStr(t, pathVal, path, A_MAX_PATH_LEN);

    AAllowBlocking();
    status = access(path, R_OK);
    AEndBlocking();
        
    return status == 0 ? ATrue : AFalse;
}


AValue AStatIsWritable(AThread *t, AValue *frame)
{
    AValue pathVal = AMemberDirect(frame[0], 0);
    char path[A_MAX_PATH_LEN];
    int status;
    
    AGetStr(t, pathVal, path, A_MAX_PATH_LEN);

    AAllowBlocking();
    status = access(path, W_OK);
    AEndBlocking();
    
    return status == 0 ? ATrue : AFalse;
}


AValue AStatOwner(AThread *t, AValue *frame)
{
    /* IDEA: Implement */
    return AMakeStr(t, "");
}


AValue AStatOwnerPerm(AThread *t, AValue *frame)
{
    /* IDEA: Implement */
    return ANil;
}

    
AValue AStatOtherPerm(AThread *t, AValue *frame)
{
    /* IDEA: Implement */
    return ANil;
}


static ABool MatchName(const char *str, const char *mask)
{
    for (; *mask != '\0'; mask++) {
        if (*mask == '*') {
            const char *s = str;
            while (*s != '\0')
                s++;
            for (; s >= str; s--) {
                if (MatchName(s, mask + 1))
                    return TRUE;
            }
            return FALSE;
        } else if (*mask == '?') {
            if (*str == '\0')
                return FALSE;
            str++;
        } else {
            if (*mask != *str)
                return FALSE;
            str++;
        }
    }

    return *str == '\0';
}


AValue AListDirImpl(AThread *t, const AMixedCValue *frame)
{
    DIR *d;
    AValue *tmp;
    AValue ret;
    struct dirent *entry;

    tmp = AAllocTemps(t, 2);

    AAllowBlocking();
    d = opendir(frame[0].strValue);
    AEndBlocking();
    
    if (d == NULL) {
        AFreeTemps(t, 2);
        return ARaiseErrnoIoError(t);
    }
    
    if (ATry(t)) {
        AAllowBlocking();
        closedir(d);
        AEndBlocking();
        return AError;
    }
    
    tmp[1] = AZero; /* IDEA: Is this superfluous? */
    tmp[0] = AMakeArray(t, 0);

    for (;;) {
        const char *n;

        AAllowBlocking();
        entry = readdir(d);
        AEndBlocking();
        
        if (entry == NULL)
            break;
        
        n = entry->d_name;
        
        if (!(n[0] == '.' && (strcmp(n, ".") == 0 || strcmp(n, "..") == 0))) {
            if (frame[1].strValue == NULL
                || MatchName(n, frame[1].strValue)) {
                tmp[1] = AMakeStr(t, n);
                AAppendArray(t, tmp[0], tmp[1]);
            }
        }
    }

    AEndTry(t);

    AAllowBlocking();
    closedir(d);
    AEndBlocking();

    ret = tmp[0];
    AFreeTemps(t, 2);
    return ret;
}


static AValue WalkDirRecursive(AThread *t, AValue *data, const char *base,
                               char *rel, const char *mask)
{
    DIR *d;
    struct dirent *entry;
    struct stat buf;
    char path[A_MAX_PATH_LEN];

    /* IDEA: Use logical try/finally to avoid leaks.. */

    strcpy(path, base);
    strcat(path, A_DIR_SEPARATOR_STRING);
    strcat(path, rel);

    AAllowBlocking();
    d = opendir(path);
    AEndBlocking();
    
    if (d == NULL)
        return ARaiseErrnoIoError(t);
    
    if (ATry(t)) {
        AAllowBlocking();
        closedir(d);
        AEndBlocking();
        return AError;
    }

    for (;;) {
        const char *n;

        AAllowBlocking();
        entry = readdir(d);
        AEndBlocking();

        if (entry == NULL)
            break;
        
        n = entry->d_name;

        if (n[0] == '.' && (strcmp(n, ".") == 0 || strcmp(n, "..") == 0))
            continue;

        strcpy(path, base);
        strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, rel);
        strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, n);

        if (stat(path, &buf) == -1) {
            closedir(d);
            return ARaiseErrnoIoError(t);
        }

        /* Concatenate n to rel and store the result in path. */
        strcpy(path, rel);
        if (*path != '\0')
            strcat(path, A_DIR_SEPARATOR_STRING);
        strcat(path, n);
        
        if (mask == NULL || MatchName(n, mask)) {
            data[1] = AMakeStr(t, path);
            AAppendArray(t, data[0], data[1]);
        }
        
        /* Descent into subdirectories. */
        if (S_ISDIR(buf.st_mode)) {
            if (AIsError(WalkDirRecursive(t, data, base, path, mask))) {
                AAllowBlocking();
                closedir(d);
                AEndBlocking();
                return AError;
            }
        }
    }

    AEndTry(t);

    AAllowBlocking();
    closedir(d);
    AEndBlocking();

    return data[0];
}


AValue AWalkDirImpl(AThread *t, const AMixedCValue *frame)
{
    AValue *tmp;
    AValue ret;

    tmp = AAllocTemps(t, 2);
    tmp[1] = AZero; /* IDEA: Is this superfluous? */
    tmp[0] = AMakeArray(t, 0);

    ret = WalkDirRecursive(t, tmp, frame[0].strValue, "", frame[1].strValue);

    AFreeTemps(t, 2);
    return ret;
}


AValue AGetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    char value[MAX_ENV_VALUE_LEN];
    DWORD ret;

    ret = GetEnvironmentVariable(frame[0].strValue, value, sizeof(value));
    if (ret > sizeof(value))
        return ARaiseByNum(t, AErrorClassNum[EX_RESOURCE_ERROR], NULL);
    else if (ret == 0) {
        if (GetLastError() == ERROR_ENVVAR_NOT_FOUND)
            return ANil;
        else {
			/* Assume that the variable is empty. This might rely on 
			   undocumented behaviour. */
            strcpy(value, "");
        }
    }    
    
    return AMakeStr(t, value);
}


AValue ASetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    if (!SetEnvironmentVariable(frame[0].strValue, frame[1].strValue))
        return ARaiseByNum(t, AErrorClassNum[EX_RESOURCE_ERROR], NULL);
    else
        return ANil;
}


AValue AUnsetEnvImpl(AThread *t, const AMixedCValue *frame)
{
    if (!SetEnvironmentVariable(frame[0].strValue, NULL)) {
        /* It is not an error to unset a nonexistent environment variable. */
        if (GetLastError() == ERROR_ENVVAR_NOT_FOUND)
            return ANil;
        return ARaiseByNum(t, AErrorClassNum[EX_RESOURCE_ERROR], NULL);
    } else
        return ANil;
}


AValue ASetPermImpl(AThread *t, const AMixedCValue *frame)
{
    /* IDEA: Implement */
    return ANil;
}


static AValue ASetFileTime(AThread *t, AValue *frame, ABool isAccess)
{
    char path[A_MAX_PATH_LEN];
    AValue v;
    double s;
    SYSTEMTIME sysTime;
    FILETIME localFileTime;
    FILETIME fileTime;
    HANDLE handle;
    ABool status;

    AGetStr(t, frame[0], path, sizeof(path));

    v = AMember(t, frame[1], "year");
    if (AIsError(v))
        return AError;
    sysTime.wYear = AGetInt(t, v);

    v = AMember(t, frame[1], "month");
    if (AIsError(v))
        return AError;
    sysTime.wMonth = AGetInt(t, v);

    v = AMember(t, frame[1], "day");
    if (AIsError(v))
        return AError;
    sysTime.wDay = AGetInt(t, v);
    sysTime.wDayOfWeek = 0;

    v = AMember(t, frame[1], "hour");
    if (AIsError(v))
        return AError;
    sysTime.wHour = AGetInt(t, v);

    v = AMember(t, frame[1], "minute");
    if (AIsError(v))
        return AError;
    sysTime.wMinute = AGetInt(t, v);

    v = AMember(t, frame[1], "second");
    if (AIsError(v))
        return AError;
    s = AGetFloat(t, v);
    sysTime.wSecond = s;
    sysTime.wMilliseconds = (s - (int)s) * 1000;
    
    if (!SystemTimeToFileTime(&sysTime, &localFileTime))
        return ARaiseValueError(t, NULL);

    if (!LocalFileTimeToFileTime(&localFileTime, &fileTime))
        return ARaiseWin32Exception(t, AStdIoErrorNum);

    AAllowBlocking();
    handle = CreateFile(path, FILE_WRITE_ATTRIBUTES,
                        FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
                        OPEN_EXISTING, 0, NULL);
    AEndBlocking();
    
    if (handle == INVALID_HANDLE_VALUE)
        return ARaiseWin32Exception(t, AStdIoErrorNum);

    AAllowBlocking();
    if (isAccess)
        status = SetFileTime(handle, NULL, &fileTime, NULL);
    else
        status = SetFileTime(handle, NULL, NULL, &fileTime);
    AEndBlocking();

    if (!status)
        return ARaiseWin32Exception(t, AStdIoErrorNum);

    CloseHandle(handle);

    return ANil;
}

    
AValue ASetModificationTime(AThread *t, AValue *frame)
{
    return ASetFileTime(t, frame, FALSE);
}


AValue ASetAccessTime(AThread *t, AValue *frame)
{
    return ASetFileTime(t, frame, TRUE);
}


AValue AListEnv(AThread *t, AValue *frame)
{
    char *env = GetEnvironmentStrings();
    char *start = env;

    int i, j, k;
    int len;

    /* env points to a concatenated list of environment strings. Each of these
       is of the form "name=value". They are separated by 0 bytes and ended
       with two 0 bytes. */

    /* Count the number of environment variables. */
    len = 0;
    for (i = 0; env[i] != '\0' || env[i + 1] != '\0'; i++) {
        /* Only count environment variables whose name is not empty. */
        if ((i == 0 || env[i - 1] == '\0') && env[i] != '=' && env[i] != '\0')
            len++;
    }

    frame[0] = AMakeArray(t, len);

    j = 0;
    i = 0;
    while (i < len) {
        char *s = env + j;
        char var[1024];
        AValue ss;

        for (k = 0; k < 1023 && s[k] != '\0' && s[k] != '='; k++)
            var[k] = s[k];
        var[k] = '\0';

        if (k > 0) {
            ss = AMakeStr(t, var);
            ASetArrayItem(t, frame[0], i, ss);
            i++;
        }

        j += k;
        while (env[j] != '\0')
            j++;
        j++;
    }

    FreeEnvironmentStrings(start);

    return frame[0];
}


AValue AOsUser(AThread *t, AValue *frame)
{
    /* See also: NetWkstaGetInfo() (NT) */

    char buf[100];
    DWORD len;

    len = sizeof(buf);
    if (WNetGetUser(NULL, buf, &len) != NO_ERROR)
        return ARaiseIoError(t, NULL);
    
    return AMakeStr(t, buf);
}


#endif
