/* unicode_module.c - unicode module

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include <stdio.h>
#include "alore.h"
#include "unicode_module.h"
#include "internal.h"


#define WORD_BUCKET_SIZE 128
#define MAX_WORD_LEN 128


static void GetUnicodeWord(int n, char *buf)
{
    int bi = n / WORD_BUCKET_SIZE;
    int wi = bi * WORD_BUCKET_SIZE;
    int ti = AUnicodeWordLUT[bi];

    strcpy(buf, "");
    for (;;) {
        int j = AUnicodeWordList[ti]; /* Shared prefix length */
        int k = 0;
        ti++;
        do {
            buf[j + k] = AUnicodeWordList[ti + k];
            k++;
        } while ((unsigned char)buf[j + k - 1] < 128);
        buf[j + k - 1] ^= 128;

        if (wi == n) {
            buf[j + k] = '\0';
            break;
        }

        ti += k;
        wi++;
    }        
}


static void GetGenericUnicodeName(int c, char *buf)
{
    int i;
    int ci;
    unsigned short w[32];

    /* Find the start location in the name list using a lookup table. */
    i = 0;
    while (AUnicodeNameLUT[i] != 0 && c >= AUnicodeNameLUT[i + 2])
        i += 3;

    ci = AUnicodeNameLUT[i];

    /* Is this an invalid character? */
    if (ci == 0 || c < ci) {
        strcpy(buf, "");
        return;
    }

    /* Iterate words in the name list until the requested name is
       reached. */
    i = AUnicodeNameLUT[i + 1];
    for (;;) {
        int j = AUnicodeNameList[i]; /* Shared prefix length */
        int k = 0;
        i++;
        
        /* Read word numbers not part of the shared prefix. The last word
           has the highest bit set. */
        do {
            w[j + k] = (AUnicodeNameList[i + k * 2] |
                        (AUnicodeNameList[i + k * 2 + 1] << 8));
            k++;
        } while (w[j + k - 1] < 32768);
        w[j + k - 1] ^= 32768;

        /* Found the requested name? */
        if (ci == c) {
            /* Build the name from component word numbers. */
            k += j;
            j = 0;
            for (i = 0; i < k; i++) {
                GetUnicodeWord(w[i], buf + j);
                j = strlen(buf);
                buf[j++] = ' ';
            }
            buf[j - 1] = '\0';
            break;
        }

        /* Advance to the next name. */
        i += k * 2;
        ci++;
    }
}


/* Definitions for processing hangul syllables */

#define HS_Start 44032
#define HS_End 55203
#define HS_N1 19
#define HS_N2 21
#define HS_N3 28

static const char * const HS_Comp1[HS_N1] = {
    "G", "GG", "N", "D", "DD", "R", "M", "B", "BB", "S", "SS", "", "J", "JJ",
    "C", "K", "T", "P", "H"
};

static const char * const HS_Comp2[HS_N2] = {
    "A", "AE", "YA", "YAE", "EO", "E", "YEO", "YE", "O", "WA", "WAE", "OE",
    "YO", "U", "WEO", "WE", "WI", "YU", "EU", "YI", "I"
};

static const char * const HS_Comp3[HS_N3] = {
    "", "G", "GG", "GS", "N", "NJ", "NH", "D", "L", "LG", "LM", "LB", "LS",
    "LT", "LP", "LH", "M", "B", "BS", "S", "SS", "NG", "J", "C", "K", "T",
    "P", "H"
};


static void GetHangulSyllableName(int ch, char *buf)
{
    int c1, c2, c3;

    ch -= HS_Start;
    c1 = ch / (HS_N2 * HS_N3);
    c2 = (ch / HS_N3) % HS_N2;
    c3 = ch % HS_N3;

    strcpy(buf, "HANGUL SYLLABLE ");
    strcat(buf, HS_Comp1[c1]);
    strcat(buf, HS_Comp2[c2]);
    strcat(buf, HS_Comp3[c3]);
}


static int FindPrefix(char *s, const char *const*a, int n)
{
    int i;
    int found = -1;

    for (i = 0; i < n; i++) {
        if (strncmp(s, a[i], strlen(a[i])) == 0) {
            if (found == -1 || strlen(a[i]) > strlen(a[found]))
                found = i;
        }
    }
    return found;
}


static int ParseHangulSyllable(char *name)
{
    int i1, i2, i3;

    i1 = FindPrefix(name, HS_Comp1, HS_N1);
    if (i1 >= 0) {
        name += strlen(HS_Comp1[i1]);
        i2 = FindPrefix(name, HS_Comp2, HS_N2);
        if (i2 >= 0) {
            name += strlen(HS_Comp2[i2]);
            i3 = FindPrefix(name, HS_Comp3, HS_N3);
            if (i3 >= 0 && strcmp(name, HS_Comp3[i3]) == 0)
                return HS_Start + i1 * (HS_N2 * HS_N3) + i2 * HS_N3 + i3;
        }
    }

    return 0;
}


#define IsCJKUnified(ch) \
    ((ch) >= 13312 && (ch) <= 40869 && !((ch) >= 19894 && (ch) <= 19967))


static void GetUnicodeName(int ch, char *buf)
{
    if (IsCJKUnified(ch))
        sprintf(buf, "CJK UNIFIED IDEOGRAPH-%.4X", ch);
    else if (ch >= HS_Start && ch <= HS_End)
        GetHangulSyllableName(ch, buf);
    else
        GetGenericUnicodeName(ch, buf);
}


static AValue UnicodeName(AThread *t, AValue *frame)
{
    int ch = AGetCh(t, frame[0]);
    char buf[128];
    
    GetUnicodeName(ch, buf);
    if (*buf == '\0')
        ARaiseValueError(t, "Invalid Unicode character");
    
    return AMakeStr(t, buf);
}


#define HASH_SIZE 17299


static int UnicodeHash(char *buf)
{
    int hash = 12345;
    while (*buf != '\0') {
        hash = (hash * 53 + *buf) % HASH_SIZE;
        buf++;
    }
    return hash;
}


static AValue LookupChr(AThread *t, AValue *frame)
{
    char buf[256];
    char name[128];
    int i;
    
    AGetStr(t, frame[0], buf, sizeof(buf));

    if (strncmp(buf, "CJK UNIFIED IDEOGRAPH-", 22) == 0) {
        char *p = buf + 22;
        if (AIsHexDigit(p[0]) && AIsHexDigit(p[1]) && AIsHexDigit(p[2]) &&
            AIsHexDigit(p[3]) && p[4] == '\0') {
            i = AUnicodeSequenceValue(p);
            if (!IsCJKUnified(i))
                goto Error;
            return AMakeCh(t, i);
        } else
            goto Error;
    } else if (strncmp(buf, "HANGUL SYLLABLE ", 16) == 0) {
        i = ParseHangulSyllable(buf + 16);
        if (i == 0)
            goto Error;
        return AMakeCh(t, i);
    }

    i = UnicodeHash(buf);
    while (AUnicodeNameHashTable[i] != 0) {
        GetUnicodeName(AUnicodeNameHashTable[i], name);
        if (strcmp(name, buf) == 0)
            return AMakeCh(t, AUnicodeNameHashTable[i]);
        i = (i + 13) % HASH_SIZE;
    }

  Error:
    
    return ARaiseValueError(t, "Invalid Unicode character name");
}


A_MODULE(unicode, "unicode")
    A_SUB("UnicodeName", 1, 0, UnicodeName)
    A_SUB("LookupChr", 1, 0, LookupChr)
A_END_MODULE()
