/* util.c - Miscellaneous helper functions

   Copyright (c) 2010 Jukka Lehtosalo

   Alore is licensed under the terms of the MIT license.  See the file
   LICENSE.txt in the distribution.
*/

#include <ctype.h>
#include "alore.h"
#include "util.h"
#include "str.h"


#define MAX_NUMBER_OF_OS_SERVICE_ARGS 4
#define MAX_OS_SERVICE_STRING_LENGTH 1024 /* FIX!! maximum of all possible
                                             strings, path len etc. */


AValue ACallOsService(AThread *t, AValue *args,
                     const AMixedCValueType *argTypes,
                     AOsServiceFunction func)
{
    AMixedCValue cargs[MAX_NUMBER_OF_OS_SERVICE_ARGS];
    char strings[MAX_NUMBER_OF_OS_SERVICE_ARGS][MAX_OS_SERVICE_STRING_LENGTH];
    int i;
    AValue retValue;
    int numArgs;

    /* Calculate the number of arguments. */
    for (numArgs = 0; argTypes[numArgs] != A_MVT_LAST_ARG; numArgs++);

    if (numArgs > MAX_NUMBER_OF_OS_SERVICE_ARGS)
        return ARaiseValueErrorND(t); /* IDEA: Raise internal error? */
    
    /* Convert and store arguments. */
    for (i = 0; i < numArgs; i++) {
        if (argTypes[i] == A_MVT_INT) {
            if (AIsShortInt(args[i]))
                cargs[i].intValue = AValueToInt(args[i]);
            else
                return ARaiseTypeErrorND(t);
            /* FIX: Handle some LongInt values as well. */
        } else if (argTypes[i] == A_MVT_FLOAT) {
            if (AIsShortInt(args[i]))
                cargs[i].floatValue = AValueToInt(args[i]);
            else if (AIsFloat(args[i]))
                cargs[i].floatValue = AValueToFloat(args[i]);
            else
                return ARaiseTypeErrorND(t);
        } else if (argTypes[i] == A_MVT_STR || argTypes[i] == A_MVT_OPT_STR) {
            char *str;

            if (args[i] != ADefault) {
                str = AAllocCStrFromString(t, args[i], strings[i],
                                          MAX_OS_SERVICE_STRING_LENGTH);
                if (str == NULL)
                    return AError;
            } else
                str = NULL;
            cargs[i].strValue = str;
        } else
            return ARaiseValueErrorND(t); /* IDEA: Raise internal error? */
    }
    
    retValue = func(t, cargs);

    for (i = 0; i < numArgs; i++) {
        if (argTypes[i] == A_MVT_STR)
            AFreeCStrFromString(cargs[i].strValue);
    }

    return retValue;
}


int AEndsWith(const char *str, const char *suffix)
{
    int sl = strlen(str);
    int pl = strlen(suffix);

    return sl >= pl && memcmp(str + sl - pl, suffix, pl) == 0;
}


#ifdef A_HAVE_WINDOWS

int APathEndsWith(const char *str, const char *suffix)
{
    int sl = strlen(str);
    int pl = strlen(suffix);
    int i;

    if (sl < pl)
        return 0;

    for (i = 1; i <= pl; i++) {
        char c1 = str[sl - i];
        char c2 = suffix[pl - i];
        if (tolower(c1) != tolower(c2) &&
            (c2 != '/' || !A_IS_DIR_SEPARATOR(c1)))
            return 0;
    }

    return 1;
}

#endif
